#include "devXGD.h"
#include "xGDprot.h"
#include "xGDtalk.h"

/* Device Driver Actions */

#if R_VERSION >= R_Version(2,7,0)
#define RCONST const
#else
#define RCONST
#endif

static void newXGD_Activate(NewDevDesc *dd);
static void newXGD_Circle(double x, double y, double r,
			  R_GE_gcontext *gc,
			  NewDevDesc *dd);
static void newXGD_Clip(double x0, double x1, double y0, double y1,
			NewDevDesc *dd);
static void newXGD_Close(NewDevDesc *dd);
static void newXGD_Deactivate(NewDevDesc *dd);
static void newXGD_Hold(NewDevDesc *dd);
static Rboolean newXGD_Locator(double *x, double *y, NewDevDesc *dd);
static void newXGD_Line(double x1, double y1, double x2, double y2,
			R_GE_gcontext *gc,
			NewDevDesc *dd);
static void newXGD_MetricInfo(int c, 
			      R_GE_gcontext *gc,
			      double* ascent, double* descent,
			      double* width, NewDevDesc *dd);
static void newXGD_Mode(int mode, NewDevDesc *dd);
static void newXGD_NewPage(R_GE_gcontext *gc, NewDevDesc *dd);
Rboolean newXGD_Open(NewDevDesc *dd, newXGDDesc *xd,
		     char *dsp, double w, double h, char *host, int port);
static void newXGD_Polygon(int n, double *x, double *y,
			   R_GE_gcontext *gc,
			   NewDevDesc *dd);
static void newXGD_Polyline(int n, double *x, double *y,
			     R_GE_gcontext *gc,
			     NewDevDesc *dd);
static void newXGD_Rect(double x0, double y0, double x1, double y1,
			 R_GE_gcontext *gc,
			 NewDevDesc *dd);
static void newXGD_Size(double *left, double *right,
			 double *bottom, double *top,
			 NewDevDesc *dd);
static double newXGD_StrWidth(RCONST char *str, 
			       R_GE_gcontext *gc,
			       NewDevDesc *dd);
static void newXGD_Text(double x, double y, RCONST char *str,
			 double rot, double hadj,
			 R_GE_gcontext *gc,
			 NewDevDesc *dd);


static R_GE_gcontext lastGC;

/* check changes in GC and issue corresponding commands if necessary */
static void checkGC(newXGDDesc *xd, R_GE_gcontext *gc) {
  if (gc->col != lastGC.col)
    sendCMD(xd, CMD_SetColor, &gc->col, 4);
  if (gc->fill != lastGC.fill)
    sendCMD(xd, CMD_SetFill, &gc->fill, 4);

  if (gc->lwd != lastGC.lwd || gc->lty != lastGC.lty) {
    struct sp_SetLinePar par;
    par.lwd = gc->lwd;
    par.lty = gc->lty;
    sendCMD(xd, CMD_SetLinePar, &par, sizeof(par));
  }

  if (gc->cex!=lastGC.cex || gc->ps!=lastGC.ps || gc->fontface!=lastGC.fontface || strcmp(gc->fontfamily, lastGC.fontfamily)) {
    struct sp_SetFont *par;
    int plen = 33 + strlen(gc->fontfamily);
    par = (struct sp_SetFont*) malloc(plen);
    par->cex = gc->cex;
    par->ps = gc->ps;
    par->lineheight = gc->lineheight;
    par->fontface = gc->fontface;
    strcpy(par->fontfamily, gc->fontfamily);
    sendCMD(xd, CMD_SetFont, par, plen);
#ifdef XGD_DEBUG
	printf(">> font changed: \"%s\", %x, %f, %f, %f\n", gc->fontfamily,
		   gc->fontface, gc->cex, gc->ps, gc->lineheight); 
#endif
  }
  memcpy(&lastGC, gc, sizeof(lastGC));
}

/* re-set the GC - i.e. send commands for all monitored GC entries */
static void sendAllGC(newXGDDesc *xd, R_GE_gcontext *gc) {
#ifdef XGD_DEBUG
	printf("Basic GC:\n col=%08x\n fill=%08x\n gamma=%f\n lwd=%f\n lty=%08x\n cex=%f\n ps=%f\n lineheight=%f\n fontface=%d\n fantfamily=\"%s\"\n\n",
		   gc->col, gc->fill, gc->gamma, gc->lwd, gc->lty,
		   gc->cex, gc->ps, gc->lineheight, gc->fontface, gc->fontfamily);
#endif
	
  sendCMD(xd, CMD_SetColor, &gc->col, 4);
  sendCMD(xd, CMD_SetFill, &gc->fill, 4);

  {
    struct sp_SetLinePar par;
    par.lwd = gc->lwd;
    par.lty = gc->lty;
    sendCMD(xd, CMD_SetLinePar, &par, sizeof(par));
  }

  {
    struct sp_SetFont *par;
    int plen = 33 + strlen(gc->fontfamily);
    par = (struct sp_SetFont*) malloc(plen);
    par->cex = gc->cex;
    par->ps = gc->ps;
    par->lineheight = gc->lineheight;
    par->fontface = gc->fontface;
    strcpy(par->fontfamily, gc->fontfamily);
    sendCMD(xd, CMD_SetFont, par, plen);
  }

  memcpy(&lastGC, gc, sizeof(lastGC));
}

static void newXGD_Activate(NewDevDesc *dd)
{

   newXGDDesc *xd = (newXGDDesc *) dd->deviceSpecific;

   sendCMD(xd, CMD_Activate, 0, 0);
}

static void newXGD_Circle(double x, double y, double r,  R_GE_gcontext *gc,  NewDevDesc *dd)
{

   newXGDDesc *xd = (newXGDDesc *) dd->deviceSpecific;
   struct sp_Circle par;

   checkGC(xd, gc);
   par.x = x;
   par.y = y;
   par.r = r;
   sendCMD(xd, CMD_Circle, &par, sizeof(par));
}

static void newXGD_Clip(double x0, double x1, double y0, double y1,  NewDevDesc *dd)
{

   newXGDDesc *xd = (newXGDDesc *) dd->deviceSpecific;
   struct sp_Clip par;

   par.x0 = x0;
   par.x1 = x1;
   par.y0 = y0;
   par.y1 = y1;
   sendCMD(xd, CMD_Clip, &par, sizeof(par));
}

static void newXGD_Close(NewDevDesc *dd)
{

   newXGDDesc *xd = (newXGDDesc *) dd->deviceSpecific;

   sendCMD(xd, CMD_Close, 0, 0);
}

static void newXGD_Deactivate(NewDevDesc *dd)
{

   newXGDDesc *xd = (newXGDDesc *) dd->deviceSpecific;

   sendCMD(xd, CMD_Deactivate, 0, 0);
}

static void newXGD_Hold(NewDevDesc *dd)
{

   newXGDDesc *xd = (newXGDDesc *) dd->deviceSpecific;

   sendCMD(xd, CMD_Hold, 0, 0);
}

static Rboolean newXGD_Locator(double *x, double *y, NewDevDesc *dd)
{

   newXGDDesc *xd = (newXGDDesc *) dd->deviceSpecific;
   struct sr_Locator ret;

   talkCMD(xd, CMD_Locator, 0, 0, &ret, sizeof(ret));
   *x = ret.x;
   *y = ret.y;
   return ret.ret;
}

static void newXGD_Line(double x1, double y1, double x2, double y2,  R_GE_gcontext *gc,  NewDevDesc *dd)
{

   newXGDDesc *xd = (newXGDDesc *) dd->deviceSpecific;
   struct sp_Line par;

   checkGC(xd, gc);
   par.x1 = x1;
   par.y1 = y1;
   par.x2 = x2;
   par.y2 = y2;
   sendCMD(xd, CMD_Line, &par, sizeof(par));
}

static void newXGD_MetricInfo(int c,  R_GE_gcontext *gc,  double* ascent, double* descent,  double* width, NewDevDesc *dd)
{

   newXGDDesc *xd = (newXGDDesc *) dd->deviceSpecific;
   struct sp_MetricInfo par;
   struct sr_MetricInfo ret;

   checkGC(xd, gc);
   par.c = c;
   talkCMD(xd, CMD_MetricInfo, &par, sizeof(par), &ret, sizeof(ret));
   
   *ascent = ret.ascent;
   *descent = ret.descent;
   *width = ret.width;
}

static void newXGD_Mode(int mode, NewDevDesc *dd)
{

   newXGDDesc *xd = (newXGDDesc *) dd->deviceSpecific;
   struct sp_Mode par;

   par.mode = mode;
   sendCMD(xd, CMD_Mode, &par, sizeof(par));
}

static void newXGD_NewPage(R_GE_gcontext *gc, NewDevDesc *dd)
{

   newXGDDesc *xd = (newXGDDesc *) dd->deviceSpecific;
   struct sp_NewPage par;

   sendCMD(xd, CMD_NewPage, &par, sizeof(par));
   /* this is an exception - we send all GC attributes just after the NewPage command */
   sendAllGC(xd, gc);
}

Rboolean newXGD_Open(NewDevDesc *dd, newXGDDesc *xd,  char *dsp, double w, double h, char *host, int port)
{

   struct sp_Open *par;
   int plen = sizeof(*par)+strlen(dsp);

   if (connectXGD(xd,host,port)) return FALSE;

   xd->fill = 0xffffffff; /* transparent, was R_RGB(255, 255, 255); */
   xd->col = R_RGB(0, 0, 0);
   xd->canvas = R_RGB(255, 255, 255);
   xd->windowWidth = w;
   xd->windowHeight = h;

   par = (struct sp_Open*) malloc(plen);
   par->w = w;
   par->h = h;
   strcpy(par->dsp, dsp);
   sendCMD(xd, CMD_Open, par, plen);

   return TRUE;
}

static void newXGD_Polygon(int n, double *x, double *y,  R_GE_gcontext *gc,  NewDevDesc *dd)
{

   newXGDDesc *xd = (newXGDDesc *) dd->deviceSpecific;
   struct sp_Polygon *par;
   int plen = sizeof(*par) + sizeof(double)*(n*2-1);
   
   checkGC(xd, gc);
   par = (struct sp_Polygon*) malloc(plen);
   par->n = n;
   memcpy(par->v, x, sizeof(double)*n);
   memcpy(par->v+n, y, sizeof(double)*n);
   sendCMD(xd, CMD_Polygon, par, plen);
}

static void newXGD_Polyline(int n, double *x, double *y,  R_GE_gcontext *gc,  NewDevDesc *dd)
{

   newXGDDesc *xd = (newXGDDesc *) dd->deviceSpecific;
   struct sp_Polyline *par;
   int plen = sizeof(*par) + sizeof(double)*(n*2-1);
   
   checkGC(xd, gc);
   par = (struct sp_Polyline*) malloc(plen);
   par->n = n;
   memcpy(par->v, x, sizeof(double)*n);
   memcpy(par->v+n, y, sizeof(double)*n);
   sendCMD(xd, CMD_Polyline, par, plen);
}

static void newXGD_Rect(double x0, double y0, double x1, double y1,  R_GE_gcontext *gc,  NewDevDesc *dd)
{

   newXGDDesc *xd = (newXGDDesc *) dd->deviceSpecific;
   struct sp_Rect par;

   checkGC(xd, gc);
   par.x0 = x0;
   par.y0 = y0;
   par.x1 = x1;
   par.y1 = y1;
   sendCMD(xd, CMD_Rect, &par, sizeof(par));
}

static void newXGD_Size(double *left, double *right,  double *bottom, double *top,  NewDevDesc *dd)
{

   newXGDDesc *xd = (newXGDDesc *) dd->deviceSpecific;
   struct sr_Size par;

   talkCMD(xd, CMD_GetSize, 0, 0, &par, sizeof(par));
   *left = par.left;
   *right = par.right;
   *bottom = par.bottom;
   *top = par.top;
}

static double newXGD_StrWidth(RCONST char *str,  R_GE_gcontext *gc,  NewDevDesc *dd)
{

   newXGDDesc *xd = (newXGDDesc *) dd->deviceSpecific;
   struct sp_StrWidth *par;
   struct sr_StrWidth ret;

   checkGC(xd, gc);
   talkCMD(xd, CMD_StrWidth, str, strlen(str)+1, &ret, sizeof(ret));
   return ret.w;
}

static void newXGD_Text(double x, double y, RCONST char *str,  double rot, double hadj,  R_GE_gcontext *gc,  NewDevDesc *dd)
{

   newXGDDesc *xd = (newXGDDesc *) dd->deviceSpecific;
   int plen = 33 + strlen(str);
   struct sp_Text *par = (struct sp_Text*) malloc(plen);
   
   checkGC(xd, gc);
   par->x = x;
   par->y = y;
   par->rot = rot;
   par->hadj = hadj;
   strcpy(par->str, str);
   sendCMD(xd, CMD_Text, par, plen);
}

void setupXGDfunctions(NewDevDesc *dd) {
    dd->close = newXGD_Close;
    dd->activate = newXGD_Activate;
    dd->deactivate = newXGD_Deactivate;
    dd->size = newXGD_Size;
    dd->newPage = newXGD_NewPage;
    dd->clip = newXGD_Clip;
    dd->strWidth = newXGD_StrWidth;
    dd->text = newXGD_Text;
    dd->rect = newXGD_Rect;
    dd->circle = newXGD_Circle;
    dd->line = newXGD_Line;
    dd->polyline = newXGD_Polyline;
    dd->polygon = newXGD_Polygon;
    dd->locator = newXGD_Locator;
    dd->mode = newXGD_Mode;
    dd->metricInfo = newXGD_MetricInfo;
#if R_GE_version >= 4
    dd->hasTextUTF8 = TRUE;
    dd->strWidthUTF8 = newXGD_StrWidth;
    dd->textUTF8 = newXGD_Text;
#else
    dd->open = newXGD_Open;
    dd->hold = newXGD_Hold;
#endif
}
