#ifndef __XGD_PROT_H__
#define __XGD_PROT_H__

/* xGD protocol definition */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include "devXGD.h"

extern void setupXGDfunctions(NewDevDesc *dd);

#define CMDA_LongPar          0x20 /* use long parameter format */
#define CMDA_RequiresResponse 0x40 /* the command awaits a reply.
				      if not set, no reply may be sent.
				      This flag is irrelevant if
				      IsResponse is set (there is never
				      a response ona response)
				    */
#define CMDA_IsResponse       0x80 /* set for response packets, i.e.
				      answers to commands with
				      RequiresRseponse set.
				   */

#define CMD_Open       (0x01)
#define CMD_Activate   (0x03)
#define CMD_Circle     (0x05)
#define CMD_Clip       (0x06)
#define CMD_Close      (0x02)
#define CMD_Deactivate (0x04)
#define CMD_Hold       (0x07)
#define CMD_Locator    (0x08|CMDA_RequiresResponse)
#define CMD_Line       (0x09)
#define CMD_MetricInfo (0x0a|CMDA_RequiresResponse)
#define CMD_Mode       (0x0b)
#define CMD_NewPage    (0x0c)
#define CMD_Polygon    (0x0d)
#define CMD_Polyline   (0x0e)
#define CMD_Rect       (0x0f)
#define CMD_GetSize    (0x10|CMDA_RequiresResponse)
#define CMD_StrWidth   (0x11|CMDA_RequiresResponse)
#define CMD_Text       (0x12)

#define CMD_SetColor   (0x13)
#define CMD_SetFill    (0x14)
#define CMD_SetFont    (0x15)
#define CMD_SetLinePar (0x16)

/* parameters for CMD_Activate */
struct sp_Activate {
};

/* parameters for CMD_Circle */
struct sp_Circle {
  double x;
  double y;
  double r;
};

/* parameters for CMD_Clip */
struct sp_Clip {
  double x0;
  double x1;
  double y0;
  double y1;
};

/* parameters for CMD_Close */
struct sp_Close {
};

/* parameters for CMD_Deactivate */
struct sp_Deactivate {
};

/* parameters for CMD_Hold */
struct sp_Hold {
};

/* parameters for CMD_Locator */
struct sp_Locator {
};

/* return for CMD_Locator */
struct sr_Locator {
  double x;
  double y;
  int ret;
};

/* parameters for CMD_Line */
struct sp_Line {
  double x1;
  double y1;
  double x2;
  double y2;
};

/* parameters for CMD_MetricInfo */
struct sp_MetricInfo {
  int c;
};

/* return for CMD_MetricInfo */
struct sr_MetricInfo {
  double ascent;
  double descent;
  double width;
};

/* parameters for CMD_Mode */
struct sp_Mode {
  int mode;
};

/* parameters for CMD_NewPage */
struct sp_NewPage {
};

/* parameters for CMD_Open */
struct sp_Open {
  double w;
  double h;
  char dsp[1];
};

/* parameters for CMD_Polygon */
struct sp_Polygon {
  int n;
  double v[1];
};

/* parameters for CMD_Polyline */
struct sp_Polyline {
  int n;
  double v[1];
};

/* parameters for CMD_Rect */
struct sp_Rect {
  double x0;
  double y0;
  double x1;
  double y1;
};

/* parameters for CMD_Size */
struct sp_Size {
};

/* return for CMD_Size */
struct sr_Size {
  double left;
  double right;
  double bottom;
  double top;
};

/* parameters for CMD_StrWidth */
struct sp_StrWidth {
  char str[1];
};

/* return for CMD_StrWidth */
struct sr_StrWidth {
  double w;
};

/* parameters for CMD_Text */
struct sp_Text {
  double x;
  double y;
  double rot;
  double hadj;
  char str[1];
};

/*--- GC-related functions, not directly from RGAPI ---*/

/* parameters for CMD_SetFont */
struct sp_SetFont {
  double cex;
  double ps;
  double lineheight;
  int fontface;
  int dummy;
  char fontfamily[1];
};

/* parameters for CMD_SetLinePar */
struct sp_SetLinePar {
  double lwd;
  int lty;
  int dummy;
};

#endif
